#!/usr/bin/env ruby

# ------------------------------------------------------------------------------
# Setup:
#   1. download title&snaps from http://www.progettosnaps.net/snapshots/
#      and place them in input/titles and input/snaps
#   2. download mame/fba dat files from https://github.com/RetroPie/RetroPie-Setup/wiki/Arcade
#      and place them in input/<mame|fba>.xml
#   3. list all roms:
#      ls mame-libretro/*.zip | sed 's/.zip//' > input/mame.zips
#      ls fba/*.zip           | sed 's/.zip//' > input/fba.zips
#   4. mkdir -p output/<mame|fba>
#
# This script does:
#   0. gem install activesupport nokogiri progressbar ferret
#      ruby build_arcade_dataset.rb <mame|fba>
#   1. read official dat/xml file                          | input/<name>.xml
#   2. read emulation station mame gamelist.xml            | input/<name>-gamelist.xml
#   3. create a new gamelist.xml with title/year/manufacturer | output/<name>-gamelist.xml
#   4. match rom name with image name and generate cover   | input/titles & input/snaps
#      images (merging title & snap)                       | output/<name>
#
# TODO: whatch the output and manually fix any eventual wrong fuzzy match
# ------------------------------------------------------------------------------

require 'yaml'
require 'active_support/all'
require 'nokogiri'
require 'progressbar'

require 'ferret' # https://github.com/dbalmain/ferret/blob/master/ruby/TUTORIAL.md
                 # http://www.rubydoc.info/gems/ferret/0.11.8.6/Ferret/Search/FuzzyQuery
# ferret examples:
#   idx = Ferret::Index::Index.new()
#   idx << "phrase"   or   idx << {id: 'foo', name: 'bar'}
#   idx.search_each(name){|id, score| puts idx[id][:id] }; nil
#   q = Ferret::Search::FuzzyQuery.new(:id, name, :min_similarity => 0.7, :prefix_length => 2)
#   idx.search_each(q){|id, score| puts idx[id][:id] + "\t" + score.to_s }; nil

PROGRESSBAR_OPTIONS = {
  title:          'steps',
  starting_at:    0,
  total:          100,
  progress_mark:  '#',
  remainder_mark: '_',
  length:         79,
  format:         '%t: %J%% [%B] %e',
}

unless %w{ mame fba }.include?(ARGV[0])
  puts "USAGE: #{File.basename __FILE__} <mame|fba>"
  exit 0
end

$emu = ARGV[0]

# build mame official dataset
repo = {}
if File.exists?("#{$emu}.dat")
  repo = YAML::load_file("#{$emu}.dat")
else
  n  = Nokogiri::XML(File.open("input/#{$emu}.xml")).css("game")
  pb = ProgressBar.create(PROGRESSBAR_OPTIONS.merge total: n.size, title: "#{$emu}.dat ")
  n.each do |m|
    pb.increment
    name  = m.attr(:name)
    descr = m.css('description' ).first.try(:content)
    anno  = m.css('year'        ).first.try(:content)
    manu  = m.css('manufacturer').first.try(:content)
    if name.present? && descr.present?
      repo[name] = {descr: descr, anno: anno, manu: manu}
    else
      puts [name, descr].join("\t")
    end
  end
  File.open("#{$emu}.dat",'w'){|f| f.puts repo.to_yaml }
end

# build current roms list
roms = {}
if File.exists?("#{$emu}.roms")
  roms = YAML::load_file("#{$emu}.roms")
else
  # read current gamelist.xml
  n  = Nokogiri::XML(File.open("input/#{$emu}-gamelist.xml")).css("game")
  pb = ProgressBar.create(PROGRESSBAR_OPTIONS.merge total: n.size, title: "#{$emu}.roms")
  n.each do |m|
    pb.increment
    name  = m.attr(:name)
    name  = m.css('path').first.try(:content).to_s.sub(/^..(.+)\.zip/, '\1')
    descr = m.css('name').first.try(:content)
    if name.present? && descr.present?
      roms[name] = descr
    else
      puts [name, descr].join("\t")
    end
  end
  # integrate any evntual missing roms
  n = File.read("input/#{$emu}.zips").split("\n")
  pb = ProgressBar.create(PROGRESSBAR_OPTIONS.merge total: n.size, title: "#{$emu}.zips")
  n.each do |m|
    pb.increment
    next if roms[m]
    roms[m] = m
  end
  # write dataset
  File.open("#{$emu}.roms",'w'){|f| f.puts roms.to_yaml }
end

#puts "#{repo.size} / #{roms.size}"

# index image names for fuzzy search
t_idx = Ferret::Index::Index.new(); Dir.chdir('input/titles'){ Dir['*'] }.each{|i| t_idx << i.sub(/\....$/, '') }; nil
s_idx = Ferret::Index::Index.new(); Dir.chdir('input/snaps' ){ Dir['*'] }.each{|i| s_idx << i.sub(/\....$/, '') }; nil

ObjectSpace.garbage_collect

ms = 0.7 # min_similarity

# write gamelist with correct names and create combined snapshots
pb = ProgressBar.create(PROGRESSBAR_OPTIONS.merge total: roms.size, title: 'building')
img_dl_dir = ($emu == 'mame' ? 'mame-libretro' : 'fba')
nxb = Nokogiri::XML::Builder.new do
  gameList {
    roms.each do |name, title|
      pb.increment
      img = "downloaded_images/#{img_dl_dir}/#{name}-image.jpg"
      
      ft, fs = nil, nil
      msg = [name]
      
      if File.exist?("input/titles/#{name}.png")
        ft = "input/titles/#{name}.png"
        msg << nil
      else
        q = Ferret::Search::FuzzyQuery.new(:id, name, :min_similarity => ms, :prefix_length => 2)
        ft = 'input/titles/' + t_idx[ t_idx.search(q).hits.first.doc ][:id] + '.png' rescue nil
        msg << (ft || '--')
      end
      
      if File.exist?("input/snaps/#{name}.png")
        fs = "input/snaps/#{name}.png"
        msg << nil
      else
        q = Ferret::Search::FuzzyQuery.new(:id, name, :min_similarity => ms, :prefix_length => 2)
        fs = 'input/snaps/' + s_idx[ s_idx.search(q).hits.first.doc ][:id] + '.png' rescue nil
        msg << (fs || '--')
      end
      
      if msg.compact.size > 1
        pb.log msg.map{|i|
          (i||'--').sub(/input.titles./,'').sub(/input.snaps./,'').sub(/.png/,'').ljust 10
        }.join(" | ")
      else
        pb.log name
      end
      
      # write game cover
      unless File.exist?(img)
        img = "output/#{$emu}/#{name}-image.jpg"

        if ft && fs
          system %Q| montage #{ft} #{fs} -tile 1x2 -geometry +0+0 -quality 80 #{img} |
        elsif fn = (ft || fs)
          system %Q| convert #{fn} -quality 80 #{img} |
        end
      end
      
      game {
        path          "./#{name}.zip"
        
        if r = repo[name]
          name        r[:descr] || title
          desc
          image       File.exist?(img) ? "/home/pi/.emulationstation/downloaded_images/#{img_dl_dir}/#{name}-image.jpg" : nil
          releasedate "#{r[:anno]}0101T000000"
          developer   r[:manu]
          publisher   r[:manu]
          genre
          players
        else
          name        title
        end
      }
    end
  }
end
File.open("output/#{$emu}-gamelist.xml",'w'){|f| f.puts nxb.to_xml.gsub(/^ +/,'') }

## check names and print problems
#notitle, tocheck = [], []
#roms.each do |name, title|
#  next if repo[name]
#
#  if name != title
#    tocheck << "#{name.ljust 15}#{title}"
#  else
#    notitle << name
#  end
#end
#puts "------- TOCHECK #{tocheck.size} ----------------------------------"; puts tocheck
#puts "------- NOTITLE #{notitle.size} ----------------------------------"; puts notitle
